
module Term.UntypedEval where

open import Prelude
open import Type
open import Term

{-# NO_POSITIVITY_CHECK #-}  -- Cheating!
data Value : Set where
  nat   : Nat → Value
  fun   : (Value → Value) → Value
  error : String → Value

Env = List (Name × Value)

lookupVar : Name → Env → Value
lookupVar x [] = error $ "Unbound variable " & x
lookupVar x ((y , v) ∷ ρ) =
  ifYes x == y then v else lookupVar x ρ

vSuc : Value → Value
vSuc (nat n)   = nat (suc n)
vSuc (fun f)   = error "Cannot apply suc to function"
vSuc (error e) = error e

vApply : Value → Value → Value
vApply (nat n)   _ = error $ "Cannot apply number " & show n
vApply (fun f)   v = f v
vApply (error e) _ = error e

eval : RawTerm → Env → Value
eval (var x)     ρ = lookupVar x ρ
eval (lit n)     ρ = nat n
eval suc         ρ = fun vSuc
eval (app t t₁)  ρ = vApply (eval t ρ) (eval t₁ ρ)
eval (lam x _ t) ρ = fun $ λ v → eval t ((x , v) ∷ ρ)

-- ex₁ : RawTerm
-- ex₁ = lam "x" nat (app suc (var "x"))
