
module TypeCheck where

open import Prelude
open import Type
open import Term
open import Term.Show
open import Term.Parse

TC : Set → Set
TC A = Either String A

fail : {A : Set} → String → TC A
fail = left

data WellTyped (Γ : Context) (e : RawTerm) : Set where
  ok : (a : Type) (t : Term Γ a) → forgetTypes t ≡ e → WellTyped Γ e

data InScope (Γ : Context) (x : Name) : Set where
  ok : (a : Type) → (x , a) ∈ Γ → InScope Γ x

iSuc : ∀ {y x a Γ} → InScope Γ x → InScope ((y , a) ∷ Γ) x
iSuc (ok a i) = ok a (suc i)

lookupVar : ∀ Γ x → TC (InScope Γ x)
lookupVar [] x = fail $ "Not in scope: " & x
lookupVar ((y  , a) ∷ Γ) x with x == y
lookupVar ((.x , a) ∷ Γ) x | yes refl = return (ok a zero)
lookupVar ((y  , a) ∷ Γ) x | no  _    = iSuc <$> lookupVar Γ x

okLam : ∀ {Γ x a e} → WellTyped ((x , a) ∷ Γ) e → WellTyped Γ (lam x a e)
-- okLam (ok b t refl) = ok (_ => b) (lam _ t) refl
okLam {x = x} {a = a} (ok b t refl) = ok (a => b) (lam x t) refl

okVar : ∀ {Γ x} → InScope Γ x → WellTyped Γ (var x)
okVar {x = x} (ok a i) = ok a (var x i) refl

checkApp : ∀ {Γ e e₁} → WellTyped Γ e → WellTyped Γ e₁ →
           TC (WellTyped Γ (app e e₁))
checkApp (ok nat s refl) (ok a t refl) =
  fail $ "Cannot apply " & show s & " of type nat to " & show t
checkApp (ok (a₁ => b) s refl) (ok a t refl) with a₁ == a
checkApp (ok (a => b) s refl) (ok .a t refl) | yes refl =
  return (ok b (app s t) refl)
checkApp (ok (a => b) s refl) (ok a₁ t refl) | no  _ =
  fail $ "Expected type " & show a & ", found type " & show a₁ & " in argument "
       & show t & " to function " & show s

infer : (Γ : Context) (e : RawTerm) → TC (WellTyped Γ e)
infer Γ (var x)     = okVar <$> lookupVar Γ x
infer Γ (lit n)     = return (ok nat (lit n) refl)
infer Γ suc         = return (ok (nat => nat) suc refl)
infer Γ (app e e₁)  =
  infer Γ e  >>= λ w →
  infer Γ e₁ >>= λ w₁ →
  checkApp w w₁
infer Γ (lam x a e) = okLam <$> infer ((x , a) ∷ Γ) e

forgetRaw : ∀ {Γ e} → WellTyped Γ e → Σ Type (λ a → Term Γ a)
forgetRaw (ok a t _) = a , t

checkString : String → TC (Σ Type (Term []))
checkString s =
  case parseTerm s of λ
  { nothing  → fail $ "Parse error on: " & s
  ; (just e) → forgetRaw <$> infer [] e }

ex₂ = checkString "λ (x : nat → nat) → λ (x : nat) → x"
